/*
Copyright 1999 - CPL (Computational Perception Lab)
This code is provided with NO SUPPORT by the (somewhat) friendly people 
in the Computational Perception Lab at the Georgia Institute of Technology. 
Send questions/comments for us to ignore to:
brostow@cc.gatech.edu or wasinee@cc.gatech.edu
*/

#ifndef CAMERAFRAME_H
#define CAMERAFRAME_H


#include <dmedia/dmedia.h>
#include <dmedia/vl.h>

#include <X11/X.h>		/* X */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <Xm/MwmUtil.h>

#include <GL/glx.h>		/* OpenGL */
#include <GL/gl.h>

class cameraframeVLMainLoop {

  /*
   * ******************************************************
   * create connection to a camera using SGI video library 
   * *****************************************************
   */

private:
  VLInfoPtr           info;
  int             keep_frame;

public:
  VLNode          src;
  VLNode          drn;
  VLControlValue  val;
  int             Viddev_f;
  int             Vidnode_f;
  int             mask;
  int             camId;
  int             init_x;
  int             init_y;
  XVisualInfo     *vi;

  VLServer        svr;
  VLPath          path;
  VLBuffer        buffer;
  char            *dataPtr;
  Display         *dpy;
  Window          window;
  GLXContext      ctx;
  int             xsize; 
  int		  ysize;
  int             sockfd;

  inline void setCameraId(int id) { camId = id; }
  inline void setDevice(int dev) { Viddev_f = dev; }
  inline void setNode(int node)  { Vidnode_f = node; }
  inline void keep() { keep_frame = 1; }

  /* Report errors */
  void
  error_exit(void)
    {
      vlPerror("cameraframe");
      exit(1);
    }

  cameraframeVLMainLoop() {
    camId = 1;
    init_x = 0;
    init_y = 0;
    Viddev_f = VL_ANY;
    Vidnode_f = VL_ANY;
    mask  = VLTransferCompleteMask | VLStreamAvailableMask;
  }

  int setupVL(void)
    {
    /* Setup VL */

    /* Connect to the Daemon */
    if ((svr = vlOpenVideo("")) == NULL) {
	fprintf(stderr, "%s: vlOpenVideo failed\n", "cameraframe");
	return 1;
    }

    /* set up a drain node in memory */
    if ((drn = vlGetNode(svr, VL_DRN, VL_MEM, VL_ANY)) == -1) {
	vlPerror("get mem drain node");
	return 1;
    }

    /* set up a source node on any video device */
    if ((src = vlGetNode(svr, VL_SRC, VL_VIDEO, Vidnode_f)) == -1) {
	vlPerror("get video source node");
	return 1;
    }

    /* Create a path using the first device that will support it */
    if ((path = vlCreatePath(svr, Viddev_f, src, drn)) == -1) {
	vlPerror("get path");
	return 1;
    }

    /* set up the hardware for and define the usage of the path */
    if (vlSetupPaths(svr, (VLPathList) & path, 1, VL_SHARE, VL_SHARE) == -1) {
	vlPerror("setup paths");
	return 1;
    }

    /* Set packing mode to RGB */

    val.intVal = VL_PACKING_RGB_8;
    vlSetControl(svr, path, drn, VL_PACKING, &val);

    //val.intVal = VL_FORMAT_RGB;
    //vlSetControl(svr, path, drn, VL_FORMAT, &val);

    /* Set screen size to fraction of normal */
    val.fractVal.numerator = 1;
    val.fractVal.denominator = 2;
    vlSetControl(svr, path, drn, VL_ZOOM, &val);
    
    /* Get the video size */
    vlGetControl(svr, path, drn, VL_SIZE, &val);
    xsize = val.xyVal.x;
    ysize = val.xyVal.y;

    /* Create and register a buffer for 1 frame */
    buffer = vlCreateBuffer(svr, path, drn, 4);
    if (buffer == NULL)
        error_exit();
    vlRegisterBuffer(svr, path, drn, buffer);
    return 0;
    } // setupVL

  int addProcessEvent(VLCallbackProc processEvent) {
    /* Set processEvent() as the callback for a transfer complete event */
    if (vlSelectEvents(svr, path, mask) == -1) {
      vlPerror("select events");
      return 1;
    }

    vlAddCallback(svr, path, mask, (VLCallbackProc) processEvent, (void *) this);
    return 0;
  } // addProcessEvent

  int startMainLoop() {
    /* Begin the data transfer */
    
    if (vlBeginTransfer(svr, path, 0, NULL) == -1) {
      vlPerror("begin transfer");
      return 1;
    }    

    keep_frame = 1;

    /* Handle event dispatching */
    vlMainLoop();
    
    return 0;
  }

  void getFrame() {
    // if(!keep_frame) freeLastFrame();
    do {
      info = vlGetNextValid(svr, buffer);
    } while (!info);
    printf(".");
    fflush(stdout);
    dataPtr = (char*) vlGetActiveRegion(svr, buffer, info);
    keep_frame = 0;
  }

  void freeLastFrame() {
    if (!keep_frame) vlPutFree(svr, buffer);
    keep_frame = 1;
  }

  ~cameraframeVLMainLoop() {
    (void) vlEndTransfer(svr, path);
    (void) vlDestroyBuffer(svr, buffer);
    vlDestroyPath(svr, path);      
    (void) vlCloseVideo(svr);
  }
};



class cameraframe: public cameraframeVLMainLoop {

  /* 
   * ******************************************************
   * Use this class, if you don't want event patcher
   * But why not?
   * ******************************************************
   */
  
private:
  VLInfoPtr           info;
  int have_frame;

public:
  int startMainLoop() {
    /* Begin the data transfer */
    
    if (vlBeginTransfer(svr, path, 0, NULL) == -1) {
      vlPerror("begin transfer");
      return 1;
    }
    
    have_frame = 0;
    
    return 0;
  }       // setupVL


  void getFrame() {
    if(have_frame==1) putFrame();
    do {
      info = vlGetNextValid(svr, buffer);
    } while (!info);
    dataPtr = (char*) vlGetActiveRegion(svr, buffer, info);
    have_frame=1;
  }

  void putFrame() {
    if(have_frame) vlPutFree(svr, buffer);
    have_frame=0;
  }


};

#endif
