/*
Copyright 1999 - CPL (Computational Perception Lab)
This code is provided with NO SUPPORT by the (somewhat) friendly people 
in the Computational Perception Lab at the Georgia Institute of Technology. 
Send questions/comments for us to ignore to:
brostow@cc.gatech.edu or wasinee@cc.gatech.edu
*/

#include "macros.h"
#include "color.h"


void colorRgbToYuv(float *color)
{
  float r = color[0];
  float g = color[1];
  float b = color[2];
  
  float y = ( 0.257*r + 0.504*g + 0.098*b) +  16.0;
  float u = (-0.148*r - 0.291*g + 0.439*b) + 128.0;
  float v = ( 0.439*r - 0.368*g - 0.072*b) + 128.0;
  
  y = LIMIT (y , 0.0, 255.0);
  u = LIMIT (u , 0.0, 255.0);
  v = LIMIT (v , 0.0, 255.0);
  
  color[0] = y;
  color[1] = u;
  color[2] = v;
}


void colorYuvToRgb(float *color)
{
  float y = color[0] - 16.0;
  float u = color[1] - 128.0;
  float v = color[2] - 128.0;

  float r = 1.164*y - 0.003*u + 1.596*v;
  float g = 1.161*y - 0.397*u - 0.814*v;
  float b = 1.162*y + 2.014*u - 0.001*v;
  
  r = LIMIT (r , 0.0, 255.0);
  g = LIMIT (g , 0.0, 255.0);
  b = LIMIT (b , 0.0, 255.0);
  
  color[0] = r;
  color[1] = g;
  color[2] = b;
}

#ifdef DIGITIZEYUV              /* if we're using YUV then identity */
void colorToYuv(float *color) { }
void colorYuvTo(float *color) { }
#else                           /* if we're using YUV, then do it */
void colorToYuv(float *color) { colorRgbToYuv(color); }
void colorYuvTo(float *color) { colorYuvToRgb(color); }
#endif

#ifdef DIGITIZEYUV              /* if we're using YUV then do it */
void colorToRgb(float *color) { colorYuvToRgb(color); }
void colorRgbTo(float *color) { colorRgbToYuv(color); }
#else                           /* if we're using RGB, then identity */
void colorToRgb(float *color) { }
void colorRgbTo(float *color) { }
#endif
