/*
Copyright 1999 - CPL (Computational Perception Lab)
This code is provided with NO SUPPORT by the (somewhat) friendly people 
in the Computational Perception Lab at the Georgia Institute of Technology. 
Send questions/comments for us to ignore to:
brostow@cc.gatech.edu or wasinee@cc.gatech.edu
*/

#include <stdio.h>
#include <stdlib.h>
#include "glXDisplay.h++"


/* ---------------------------------------------------------------------- */
static int  xioerror(Display * dpy)
{
  return(-1);
}

/* ---------------------------------------------------------------------- */
static Bool waitForMapNotify(Display * dpy, XEvent * ev, XPointer arg)
{
  return (ev->type == MapNotify && ev->xmapping.window == *((Window *) arg));
}


void glXDisplay::setupWindow(int width, int height, int ncanvas) {
  /* set up the OpenGL/X window */
  int             visualAttr[] =
  {GLX_RGBA, GLX_RED_SIZE, 8,
   GLX_GREEN_SIZE, 8, GLX_BLUE_SIZE, 8,
   None};
  
  glXDisplay::ncanvas = ncanvas;
  xsize = ncanvas*width;
  ysize = height;
  init_x = 0;
  init_y = 0;
  frameXsize = width; // calculate width for display

  CreateWindowAndContext(&window, &ctx, visualAttr, "vtos:");
  XSelectInput(dpy, window, ButtonPressMask | KeyPressMask);
  glLoadIdentity();
  glViewport(0, 0, xsize, ysize);
  glOrtho(0, xsize, 0, ysize, -1, 1);
  glDrawBuffer(GL_FRONT);	/* On Vino, we only use the front buffer */
  /* video is upside down to graphics! */

  // NOTE: Comment out the following to flip image (and change drawPixels too):
  if (!flip) {
    glPixelZoom(1.0, -1.0);
  }
}

void glXDisplay::CreateWindowAndContext(Window * win, GLXContext * ctx, int *visualAttr, char name[])
{
  XSizeHints      hints;
  XSetWindowAttributes swa;
  Colormap        cmap;
  int             isRgba;
  XEvent          event;
  int             i;
  
  hints.min_width = xsize / 2;
  hints.max_width = 2 * xsize;
  hints.base_width = hints.width = xsize;
  hints.min_height = ysize / 2;
  hints.max_height = 2 * ysize;
  hints.base_height = hints.height = ysize;
  hints.flags = USSize | PMinSize | PMaxSize;
  
  /* If user requested initial window position of 0,0 */
  /* get a connection */
  
  dpy = XOpenDisplay(NULL);
  
  if (dpy == NULL) {
    fprintf(stderr, "Can't connect to display `%s'\n",
		getenv("DISPLAY"));
    exit(EXIT_FAILURE);
  }
  
  XSetIOErrorHandler(xioerror);
  
  /* get an appropriate visual */
  fprintf(stderr, "before glXChoosevisual");
  fprintf(stderr, "  %d  %d\n", DefaultScreen(dpy), visualAttr[2] );
  vi = glXChooseVisual(dpy, DefaultScreen(dpy), visualAttr);
  if (vi == NULL) {
    fprintf(stderr, "No matching visual on display `%s'\n",
	    getenv("DISPLAY"));
    exit(EXIT_FAILURE);
  }
  (void) glXGetConfig(dpy, vi, GLX_RGBA, &isRgba);
      /* create a GLX context */
  if ((*ctx = glXCreateContext(dpy, vi, 0, GL_TRUE)) == NULL) {
    fprintf(stderr, "Cannot create a context.\n");
    exit(EXIT_FAILURE);
  }
  /* create a colormap (it's empty for rgba visuals) */
  cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen),
			 vi->visual, isRgba ? AllocNone : AllocAll);
  if (cmap == NULL) {
	fprintf(stderr, "Cannot create a colormap.\n");
	exit(EXIT_FAILURE);
  }
  /* fill the colormap with something simple */
  if (!isRgba) {
    static char    *cnames[] =
    {
		"black", "red", "green", "yellow", "blue", "magenta", "cyan", "white"
    };
    
    for (i = 0; i < sizeof (cnames) / sizeof (cnames[0]); i++) {
      XStoreNamedColor(dpy, cmap, cnames[i], i, DoRed | DoGreen | DoBlue);
    }
  }
  /* create a window */
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = StructureNotifyMask | KeyPressMask | ExposureMask;
  
  *win = XCreateWindow(dpy, RootWindow(dpy, vi->screen),
		       init_x, init_y, xsize, ysize,
			   0, vi->depth, InputOutput, vi->visual,
		       CWBorderPixel | CWColormap | CWEventMask, &swa);
    
  if (*win == NULL) {
    fprintf(stderr, "Cannot create a window.\n");
    exit(EXIT_FAILURE);
  }
  
  XStoreName(dpy, *win, name);
  XSetNormalHints(dpy, *win, &hints);
  
  XMapWindow(dpy, *win);
  XIfEvent(dpy, &event, waitForMapNotify, (XPointer) win);
  
  /* Connect the context to the window */
  if (!glXMakeCurrent(dpy, *win, *ctx)) {
    fprintf(stderr, "Can't make window current to context\n");
    exit(EXIT_FAILURE);
  }
}

void glXDisplay::drawPixels(unsigned char *data, int canvasid)
{
  float startx = canvasid * (float) frameXsize;
  // glPixelZoom(1.0, -1.0);

  // NOTE: Change (ysize-1) to 0 to flip output:
  if (!flip) 
    glRasterPos2f(startx, ysize-1);
  else
    glRasterPos2f(startx, 0);

  glDrawPixels(frameXsize, ysize, GL_ABGR_EXT, GL_UNSIGNED_BYTE, data);
}
