/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

/* iveImage.c */
/* RCS: $Header: /v/projects/pfinder/src/RCS/iveImage.c,v 9.0 1997/03/08 15:56:57 cwren Exp $ */

#include "iveImage.h"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class ivePixel
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

/* these functions are return by value, whereas the macro is writable */

unsigned char
ivePixelGetColor( ivePixel *pix, ivePixelCh ch )
{
  return ( ivePixelColor( pix, ch ) );
}

unsigned char
ivePixelGetRed( ivePixel *pix )
{
  return ( pix->red );
}

unsigned char
ivePixelGetGreen( ivePixel *pix )
{
  return ( pix->green );
}

unsigned char
ivePixelGetBlue( ivePixel *pix )
{
  return ( pix->blue );
}

void
ivePixelPutColor( ivePixel *pix, ivePixelCh ch, unsigned char val )
{
  ivePixelColor( pix, ch ) = val;
}

void
ivePixelPutRed( ivePixel *pix, unsigned char val )
{
  pix->red = val;
}

void
ivePixelPutGreen( ivePixel *pix, unsigned char val )
{
  pix->green = val;
}

void
ivePixelPutBlue( ivePixel *pix, unsigned char val )
{
  pix->blue = val;
}

#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * class iveImage
 *
 * iveImageMake
 * iveImageMakeCopy
 * iveImageFree
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

iveImage *
iveImageMake( int height, int width )
{
  iveImage *im = ( iveImage * ) malloc( sizeof( iveImage ) );

  if ( im ) 
    {
      im -> height = height;
      im -> width = width;
      im -> pixelCount = width * height;
      im -> byteCount = im->pixelCount * iveBPP;
      im -> data = (unsigned char *) malloc( im->byteCount );
      if ( ! im->data )
	{
	  free( im );
	  perror("iveImageMake failed image malloc");
	  return 0;
	}
      return im;
    }
  else
    {
      perror("iveImageMake failed structure malloc");
      return (0);
    }
}

iveImage *
iveImageCopy( iveImage *src, iveImage *dst )
{
  if ( dst && src ) 
    {
      dst -> height = src->height;
      dst -> width = src->width;
      dst -> pixelCount = src->pixelCount;
      dst -> byteCount = src->byteCount;
      bcopy( src->data, dst->data, dst->byteCount );
    }
  return dst;
}

iveImage *
iveImageMakeCopy( iveImage *src )
{
  if (src)
    {
      iveImage *im = ( iveImage * ) malloc( sizeof( iveImage ) );

      if ( im ) 
	{
	  im -> height = src->height;
	  im -> width = src->width;
	  im -> pixelCount = src->pixelCount;
	  im -> byteCount = src->byteCount;
	  im -> data = (unsigned char *) malloc( im->byteCount );
	  if ( ! im->data )
	    {
	      free( im );
	      perror("iveImageMakeCopy failed image malloc");
	      return 0;
	    }
	  bcopy( src->data, im->data, im->byteCount );
	  return im;
	}
      else
	{
	  perror("iveImageMakeCopy failed structure malloc");
	  return (0);
	}
    }
  else
    {
      perror("iveImageMakeCopy no source image");
      return (0);
    }
}

iveImage *
iveImageInit( iveImage *im, int height, int width, unsigned char *buffer )
{
  im -> height = height;
  im -> width = width;
  im -> pixelCount = width * height;
  im -> byteCount = im->pixelCount * iveBPP;
  im -> data = buffer;
  return im;
}


/* this should only be called on Make or MakeCopy results */
void iveImageFree(iveImage *im)	
{
  free(im->data);
  free(im);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * iveImageReadRaw
 * iveImageWriteRaw
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

int
iveImageReadRaw( iveImage *im, FILE *stream )
{
  unsigned char rgb[3];
  int num;
  int i;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      num = fread( rgb, sizeof(unsigned char), 3, stream );
      if( EOF == num )
	return 1;
      if ( num != 3 )
	{
	  fprintf( stderr,
		  "iveImage: EOF or read raw error (%d, %d) \n", i, num );
	  return 1;
	}
      iveImageColorI( im, i, iveImageRed ) = rgb[0];
      iveImageColorI( im, i, iveImageGreen ) = rgb[1];
      iveImageColorI( im, i, iveImageBlue ) = rgb[2];
    }
  return 0;
}

int
iveImageWriteRaw( iveImage *im, FILE *stream )
{
  int i;
  float color[3];
  unsigned char rgb[3];
  unsigned char y, u, v;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      putc( iveImageColorI( im, i, iveImageRed ), stream );
      putc( iveImageColorI( im, i, iveImageGreen ), stream );
      putc( iveImageColorI( im, i, iveImageBlue ), stream );
    }
  return 0;
}

#else

/* YUV */
#include "macros.h"

int
iveImageReadRaw( iveImage *im, FILE *stream )
{
  extern void bsColorRgbTo(float *);
  float color[3];
  unsigned char rgb[3];
  unsigned char y, u, v;
  unsigned char *ucp = im->data;
  int num;
  int i;
  
  for( i = 0; i < im->pixelCount; i++ )
    {
      num = fread( rgb, sizeof(unsigned char), 3, stream );
      if( feof( stream ) )
	return 1;
      if ( num != 3 )
	{
	  fprintf( stderr, "iveImage: read raw error (%d, %d) \n", i, num );
	  return 2;
	}
      color[0] = (float) rgb[0];
      color[1] = (float) rgb[1];
      color[2] = (float) rgb[2];
      bsColorRgbTo(color);
      y = (int) color[0];
      u = (int) color[1];
      v = (int) color[2];

      ucp[YIndex(i)] = y;
      ucp[UIndex(i)] = u;
      ucp[VIndex(i)] = v;
    }
  return 0;
}

int
iveImageWriteRaw( iveImage *im, FILE *stream )
{
  extern void bsColorYuvToRgb(float *);
  int i;
  unsigned char rgb[3];
  float color[3];
  unsigned char *ucp = im->data;
  
  for( i = 0; i < im->pixelCount ; i ++ )
    {
      ivePixel ivepixel;
      iveImageGetPixelI( im, i, &ivepixel );

      color[0] = ucp[YIndex(i)];
      color[1] = ucp[UIndex(i)];
      color[2] = ucp[VIndex(i)];

      bsColorToRgb(color);
      
      rgb[0] = color[0];
      rgb[1] = color[1];
      rgb[2] = color[2];

      fwrite( rgb, sizeof(unsigned char), 3, stream );

    }
  return 0;
}

#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * iveImageGetPixel
 * iveImagePutPixel
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef DIGITIZEYUV

ivePixel *
iveImageGetPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  unsigned char *ucp = im->data + iveImageIndex( im, y, x, 0 );
  pix->alpha = *ucp++;
  pix->blue = *ucp++;
  pix->green = *ucp++;
  pix->red = *ucp++;
}

ivePixel *
iveImageGetPixelI( iveImage *im, int index, ivePixel *pix )
{
  unsigned char *ucp = im->data + index * iveBPW;
  pix->alpha = *ucp++;
  pix->blue = *ucp++;
  pix->green = *ucp++;
  pix->red = *ucp++;
}

void
iveImagePutPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  unsigned char *ucp = im->data + iveImageIndex( im, y, x, 0 );
  *ucp++ = pix->alpha;
  *ucp++ = pix->blue;
  *ucp++ = pix->green;
  *ucp++ = pix->red;
}

#else

ivePixel *
iveImageGetPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  pix -> y = iveImageColor( im, y, x, iveImageY );
  pix -> u = iveImageColor( im, y, x, iveImageU );
  pix -> v = iveImageColor( im, y, x, iveImageV );
}

ivePixel *
iveImageGetPixelI( iveImage *im, int index, ivePixel *pix )
{
  pix -> y = iveImageColorI( im, index, iveImageY );
  pix -> u = iveImageColorI( im, index, iveImageU );
  pix -> v = iveImageColorI( im, index, iveImageV );
}

void
iveImagePutPixel( iveImage *im, int y, int x, ivePixel *pix )
{
  iveImageColor( im, y, x, iveImageY ) = pix->y;
  iveImageColor( im, y, x, iveImageU ) = pix->u;
  iveImageColor( im, y, x, iveImageV ) = pix->v;
}

#endif
