/*
Copyright 1999 - CPL (Computational Perception Lab)
This code is provided with NO SUPPORT by the (somewhat) friendly people 
in the Computational Perception Lab at the Georgia Institute of Technology. 
Send questions/comments for us to ignore to:
brostow@cc.gatech.edu or wasinee@cc.gatech.edu
*/

/*
 * vtos.c - video to screen
 */
#include <sys/types.h>
#include <sys/resource.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <bstring.h>
#include <sys/time.h>
#include <limits.h>
#include <dmedia/dmedia.h>
#include <dmedia/vl.h>

#include <X11/X.h>		/* X */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <Xm/MwmUtil.h>

#include <GL/glx.h>		/* OpenGL */
#include <GL/gl.h>

extern "C" {
  #include <dmedia/moviefile.h>
  #include "iveImage.h"
  #include "color.h"
	   }

#include "glXDisplay.h++"
#include "cameraframe.h++"
#include "rgbmtrx.h++"
#include "macros.h"
#include "matlib.h"

#define FLIP_OUTPUT TRUE

#define at(tex,x,y,c) (*((tex)->data + 3*((x)+(y)*(tex)->xsize)+c))

/* GLOBALS */
glXDisplay      glx_display( FLIP_OUTPUT );

static int             readRawMode = 0;
static int             readMovMode = 0;
static int             nframes = 0;
static char            *_progName;
static int             Debug_f      = 0;
static int             reportTiming = 0;

Texture *cow;

/* ---------------------------------------------------------------------- */
int             ntoi(char *str, int *val);
void            usage(void);
/* ---------------------------------------------------------------------- */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 *
 * howslow: 
 * calculate the time it takes to process and gather NFRAMES
 *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define NFRAMES       25
#define NUM_PANELS_IN_DISPLAY 1

static int total = -1;
static int count = 0;
static struct timeval start;

static void howslow() {
  struct timeval end;
  float rate;
  float time_sec;

  total++;
  // count = number of calls since init, *including* this one
  // Thus, when count == NFRAMES, the time elapsed is for NFRAMES
  // calls to processCamEvent()

  if (count==0)  // only happens once
    {
       gettimeofday(&start);
       count = 1;
       return;
     } 
  else if ( count++ == NFRAMES ) 
    {
      gettimeofday( &end );
      // time_sec = ((float) abs(end.tv_usec - start.tv_usec))/100000.0 + 
      //  (float) abs(end.tv_sec - start.tv_sec);
      time_sec = (float) (end.tv_sec - start.tv_sec);
      rate = ((float) NFRAMES) / time_sec; // frames per second
      fprintf ( stderr, "%d [%2.2f Hz, (%d frames in %2.2f sec), total %d]\n", nframes, rate, NFRAMES, time_sec, total );
      start.tv_sec = end.tv_sec;
      start.tv_usec = end.tv_usec;
      count = 1;
      return;
    }
}


/*#####################################################################*/
////////////////////////////////////////////////////////////////////////////
//
//   processFrame: Process each frame of either camera or file input.
//
////////////////////////////////////////////////////////////////////////////


typedef struct {
  int r;
  int g;
  int b;
} IPIXEL;

PIXEL i2p(IPIXEL p) {
  PIXEL q;
  q.r=p.r;
  q.g=p.g;
  q.b=p.b;
  return q;
}
IPIXEL p2i(PIXEL p) {
  IPIXEL q;
  q.r=p.r;
  q.g=p.g;
  q.b=p.b;
  return q;
}

float mag(IPIXEL p) {
  return sqrt(p.r*p.r+p.g*p.g+p.b*p.b);
}
float dot(IPIXEL p, IPIXEL q) {
  return p.r*q.r+p.g*q.g+p.b*q.b;
}

IPIXEL sub(IPIXEL p, IPIXEL q) {
  IPIXEL c;
  c.r=p.r-q.r;
  c.g=p.g-q.g;
  c.b=p.b-q.b;
  return c;
}

IPIXEL add(IPIXEL p, IPIXEL q) {
  IPIXEL c;
  c.r=p.r+q.r;
  c.g=p.g+q.g;
  c.b=p.b+q.b;
  return c;
}

IPIXEL scale(IPIXEL p, float s) {
  IPIXEL c;
  c.r=p.r*s;
  c.g=p.g*s;
  c.b=p.b*s;
  return c;
}


void processFrame( struct mtrx newFrame ) {
  
int i,j,k,a,b,top,bottom,rr,gg,bb;
int col[320];
int row[240];
float t=0;
PIXEL p,pp;
  //  TODO: Add your processing code here.

static float wtarray[240][320];
static PIXEL newarray[240][320];
static short medianarray[3][240][320][32];

static notfirst=0;
static parity=0;
static counter=0;

/*
parity=(parity+1)%2;
*/
  for(i=0;i<240;i++)
    for(j=0;j<320;j++)
 {
   newarray[i][j]=newFrame.data[i*newFrame.width+j];
}


  /* Write the data to the screen */

for(i=0;i<240;i++)
  for(j=0;j<320;j++) {
    p=newFrame.data[i*newFrame.width+j];
    if (p.r>p.g && p.b<50 && p.r>30) {
    medianarray[0][i][j][p.r/8]++;
    medianarray[1][i][j][p.g/8]++;
    medianarray[2][i][j][p.b/8]++;
    }
    else {
    medianarray[0][i][j][p.r/8]+=2;
    medianarray[1][i][j][p.g/8]+=2;
    medianarray[2][i][j][p.b/8]+=2;
    }
}


for(i=0;i<240;i++)
  for(j=0;j<320;j++) {
    a=0;
    for(k=0;k<32;k++) a+=medianarray[0][i][j][k];
    a=a/2;
    b=0;
    for(k=0;b<a;k++) b+=medianarray[0][i][j][k];
    pp.r=8*(k-1)+4;
    a=0;
    for(k=0;k<32;k++) a+=medianarray[1][i][j][k];
    a=a/2;
    b=0;
    for(k=0;b<a;k++) b+=medianarray[1][i][j][k];
    pp.g=8*(k-1)+4;
    a=0;
    for(k=0;k<32;k++) a+=medianarray[2][i][j][k];
    a=a/2;
    b=0;
    for(k=0;b<a;k++) b+=medianarray[2][i][j][k];
    pp.b=8*(k-1)+4;
    p=newFrame.data[i*newFrame.width+j];
    if ((t=(mag(sub(p2i(pp),p2i(p)))))<45 && (p.r<2*p.g || p.b>20 || p.r<100)) {
      newFrame.data[i*newFrame.width+j].r=0;
      newFrame.data[i*newFrame.width+j].g=0;
      newFrame.data[i*newFrame.width+j].b=0;
    }
        if (p.r<p.g || p.b>50 || p.r<20) {
      newFrame.data[i*newFrame.width+j].r=0;
      newFrame.data[i*newFrame.width+j].g=0;
      newFrame.data[i*newFrame.width+j].b=0;
      }
  }
  for(k=0;k<240;k++) row[k]=0;     
  for(k=0;k<320;k++) col[k]=0;     
  for(i=0;i<240;i++)
    for(j=0;j<320;j++) 
      {
    p=newFrame.data[i*newFrame.width+j];
      	if (p.r!=0 || p.g!=0 || p.b!=0) row[i]++;
      	if (p.r!=0 || p.g!=0 || p.b!=0) col[j]++;
      }
    for(k=0;k<240;k++) a+=row[k];
    a=a/2;
    b=0;
    for(k=0;b<a;k++) b+=row[k];
    i=k-1;
    for(k=0;k<320;k++) a+=col[k];
    a=a/2;
    b=0;
    for(k=0;b<a;k++) b+=col[k];
    j=k-1;
      p.r=205;p.g=255;p.b=50;


for(a=0;a<cow->xsize/2;a++)
  for(k=0;k<cow->ysize/2;k++){
    if (at(cow,2*a,2*k,0)!=177) {
    newarray[i-k+20][-a+j].r=at(cow,2*a,2*k,0);   
 newarray[i-k+20][-a+j].g=at(cow,2*a,2*k,1);  
  newarray[i-k+20][-a+j].b=at(cow,2*a,2*k,2);
    }
  }

      /*
    for(k=i-10;k<i+10;k++) 
      newarray[k][j]=p;
    for(a=j-10;a<j+10;a++) {
    newarray[i][a]=p;*/
      /*      newarray[j+k][i+a]=p;
      newarray[j-k][i+a]=p;
      newarray[j+k][i+a]=p;*/
      /*      }*/

  for(i=0;i<240;i++)
    for(j=0;j<320;j++)
      {
	     	newFrame.data[i*newFrame.width+j]=newarray[i][j];
      } 
  /*  for(i=150;i<170;i++)
    for(j=100;j<140;j++)
      {
	newFrame.data[j*newFrame.width+i]=p;
      } 
  */
     
 counter++;
     /*if (!parity)*/  glx_display.drawPixels((unsigned char *) newFrame.data, 0);
}
/*#####################################################################*/


// Come here to set up the image matrix if using live input (camera).
void processCamEvent(VLServer svr, VLEvent * ev, void *obj)
{
    cameraframeVLMainLoop  *cam_obj = (cameraframeVLMainLoop *) obj;

    if (cam_obj == NULL) {
      fprintf(stderr, "processCamEvent: no camera presented\n");
      exit(-1);
    }

    switch (ev->reason) {
    case VLTransferComplete:

      if (reportTiming) howslow();
      
      cam_obj->getFrame();
      // -------------------------------------------------
      static struct mtrx mtrxLiveFrame;
      
      mtrxLiveFrame.width = cam_obj->xsize;
      mtrxLiveFrame.height = cam_obj->ysize;
      mtrxLiveFrame.num_elems = (cam_obj->xsize) * (cam_obj->ysize);
      mtrxLiveFrame.data = (ARGB *) cam_obj->dataPtr;
      
      processFrame( mtrxLiveFrame );

      // -------------------------------------------------
      vlPutFree(cam_obj->svr, cam_obj->buffer);

      break;


    case VLStreamAvailable:
      /* Restart the transfer because something went wrong (hickup).*/
      fprintf(stderr, "case VLStreamAvailable");

      vlBufferReset(cam_obj->svr, cam_obj->buffer);
      vlSetupPaths(cam_obj->svr, &(cam_obj->path), 1, VL_SHARE, VL_SHARE);
      vlBeginTransfer(cam_obj->svr, cam_obj->path, 0, NULL);
      break;
    }
}

// Come here to set up the image matrix if using movie files.
void processRawMovieFrame( iveImage *frame ) {
  static struct mtrx mtrxMovieFrame;

  mtrxMovieFrame.width = frame->width;
  mtrxMovieFrame.height = frame->height;
  mtrxMovieFrame.num_elems = frame->width * frame->height;
  mtrxMovieFrame.data = (ARGB *) frame->data;
  
  processFrame( mtrxMovieFrame );
}

void processMovieFrames( MVid theTrack, int firstFrame, int numFrames )
{
  DMparams     *params          = mvGetParams( theTrack );
  unsigned char *readBuff;
  int           buffsize;
  static struct mtrx mtrxMovieFrame;

  dmParamsSetEnum( params,
		   DM_IMAGE_PACKING,
		   DM_IMAGE_PACKING_XBGR );
  dmParamsSetEnum( params,
		   DM_IMAGE_INTERLACING,
		   DM_IMAGE_NONINTERLACED );
  dmParamsSetEnum( params,
		   DM_IMAGE_ORIENTATION,
		   DM_IMAGE_BOTTOM_TO_TOP );

  buffsize = dmImageFrameSize( params );

  readBuff = (unsigned char *) malloc(buffsize);

  if ( readBuff == NULL ) {
    fprintf( stderr, "Unable to allocate image buffer.\n");
    exit( EXIT_FAILURE );
  }

  mtrxMovieFrame.width = mvGetImageWidth( theTrack );
  mtrxMovieFrame.height = mvGetImageHeight( theTrack );
  mtrxMovieFrame.num_elems = mtrxMovieFrame.width * mtrxMovieFrame.height;
  
  for ( int i = 0; i < numFrames; i++ ) {
    
    if ( mvReadFrames ( theTrack, firstFrame + i, 1,
			buffsize, readBuff ) != DM_SUCCESS ) {
      fprintf( stderr, "Could not read frame %d from.", firstFrame + i);
      exit( EXIT_FAILURE );
    }
    mtrxMovieFrame.data = (ARGB *) readBuff;
    
    processFrame( mtrxMovieFrame );
  }
}


//display// static ARGB *display;


void initialize_myglobal(ARGB *dest, cameraframeVLMainLoop *cam_obj) {
  //global_frame_buffer = new ARGB[ (cam_obj->xsize * cam_obj->ysize) ];
  //calloc( (cam_obj->xsize * cam_obj->ysize),  sizeof(ARGB) );
}


/* ---------------------------------------------------------------------- */

int main(int ac, char *argv[])
{

  int            c, errflg = 0;

  _progName = argv[0];

  setpriority(PRIO_PROCESS,0,-20);
  
  cameraframeVLMainLoop *cam1 = new cameraframeVLMainLoop();

  int xsize = 320;
  int ysize = 240;

  FILE *rawfp;
  char* movFileName;

  cow=load_texture("habinnismall.ppm");
  if (cow==NULL) {fprintf(stderr,"Can't find cow.\n");return 1;}
  while ((c = getopt(ac, argv, "Dtw:h:m:r:")) != -1) {
    switch (c) {
    case 'r':    // Read a raw file.
      readRawMode++;
      rawfp = fopen(optarg, "r");
      break;
    case 'm':    // Read a movie from a file.
      readMovMode++;
      movFileName = optarg;
      break;
    case 'D':
      Debug_f++;
      break;
    case 't':
      reportTiming++;
      break;
    case 'w':
      if (ntoi(optarg, &xsize))
        errflg++;
      break;
    case 'h':
      if (ntoi(optarg, &ysize))
        errflg++;
      break;
    case '?':
      errflg++;
      break;
    }
  }
  
  if (Debug_f) {
    fprintf(stderr, "%s\n", _progName);
  }
  
  if (errflg) {
    usage();
    return 1;
  }

  /* initialize mapping */
  if (readMovMode) {           // read from a movie file
 
    int openFlag = O_RDONLY;
    MVid           theEditMovie;
    if ( mvOpenFile( movFileName, 
		     openFlag, 
		     &theEditMovie )  !=  DM_SUCCESS ) {
      fprintf( stderr, "Could not open movie file\n");
      exit( EXIT_FAILURE );
    }
    
    if ( mvFindTrackByMedium( theEditMovie, DM_IMAGE, 
			      &theEditMovie ) != DM_SUCCESS ) {
      fprintf( stderr, "Could not find track\n");
      exit( EXIT_FAILURE );
    }
    
    xsize = mvGetImageWidth( theEditMovie );
    ysize = mvGetImageHeight( theEditMovie );
    
    //display//    display = new ARGB[xsize*ysize];
    
    glx_display.setupWindow( xsize, 
			     ysize, 
			     NUM_PANELS_IN_DISPLAY );
    while(1) {
      
      processMovieFrames(theEditMovie, 1, 100);
      
      //      processRawMovieFrame( rawIm );
      
    } 
  }
  else if (readRawMode) {           // read from a raw file
    
    iveImage *rawIm = iveImageMake( ysize, xsize );
    
    //display//    display = new ARGB[xsize*ysize];
    iveImage *out = new iveImage;
    
    glx_display.setupWindow( xsize, 
			     ysize, 
			     NUM_PANELS_IN_DISPLAY );
    
    while(1) {
      
      if (iveImageReadRaw( rawIm, rawfp ))
	// read until we reach the end of file
	{
	  fprintf(stderr,"end of input\n");
	  // clean up
	  delete rawIm;
	  return 0;
	} else {
	  
	  processRawMovieFrame( rawIm );
	  
	  /* dump frames to stdout */
	  //	      iveImageWriteRaw(out, stdout);  
	}
    }
  } else {              // Get input from camera.
    /* Setup VL */
    if (cam1->setupVL())
      exit(-1);
    /* Now we can access cam1 comfortably */
    
    glx_display.setupWindow( cam1->xsize, 
			     cam1->ysize,
			     NUM_PANELS_IN_DISPLAY );
    
    
    if (cam1->addProcessEvent((VLCallbackProc) processCamEvent))
      exit(-1);
    
    fprintf(stderr,
	    "Connect to a camera: %dx%d\n", cam1->xsize, cam1->ysize);  
    
    ARGB *buffer = new ARGB[cam1->xsize*cam1->ysize];
    
    bzero(buffer, cam1->xsize*cam1->ysize*sizeof(ARGB));
    
    initialize_myglobal(buffer, cam1);
    
    cam1->startMainLoop();
    
    delete cam1;
    
    return 0;
  }
}

/* ---------------------------------------------------------------------- */

void
usage()
{
    fprintf(stderr, "usage: [options]\n", _progName);
    fprintf(stderr, "\t-R\n");
    fprintf(stderr, "\t  read sequence of images from a file\n");
    fprintf(stderr, "\t-w width\n");
    fprintf(stderr, "\t-h height\n");
    fprintf(stderr, "\t-t\n");
    fprintf(stderr, "\t  if you want to see how fast your program is\n");      
    fprintf(stderr, "\t-D\n");
    fprintf(stderr, "\t  switch to Debug mode\n");
}


int
ntoi(char *str, int *val)
{
    char           *strp;
    long            ret;

    if (*str == '\'') {
	str++;
	return (*str) ? *str : -1;
    }
    ret = strtol(str, &strp, 0);

    if (((ret == 0) && (strp == str)) ||
	    (((ret == LONG_MIN) || ret == LONG_MAX) && (errno == ERANGE)))
	return -1;

    if ((ret > INT_MAX) || (ret < INT_MIN)) {
	return -1;
    }
    *val = (int) ret;

    return 0;
}

/* ---------------------------------------------------------------------- */
