/*
Copyright 1999 - CPL (Computational Perception Lab)
This code is provided with NO SUPPORT by the (somewhat) friendly people 
in the Computational Perception Lab at the Georgia Institute of Technology. 
Send questions/comments for us to ignore to:
brostow@cc.gatech.edu or wasinee@cc.gatech.edu
*/

#ifndef RGBMTRX
#define RGBMTRX

#include "macros.h"
#include <math.h>

/* ------------------------ Type definition  ---------------------------- */
typedef char PIXEL_ARRAY[4];

typedef struct pixel {
  unsigned char a;
  unsigned char b;
  unsigned char g;
  unsigned char r;
} PIXEL;

typedef struct realrgb {
  float r;
  float g;
  float b;
} realRGB;

typedef PIXEL     MTRX_TYPE;
typedef PIXEL     ARGB;
typedef int       MTRX_IT;

typedef struct mtrx {
  MTRX_IT width,height;
  MTRX_IT num_elems;
  MTRX_TYPE* data;
} MTRX;

typedef int (*MTRXoprtFunc) (PIXEL *dest, PIXEL *src, void *data);

inline void setPixelARGB(ARGB *dest, int r, int g, int b) {
  dest->r = r;
  dest->g = g;
  dest->b = b;
}

inline void getPixelARGB(ARGB *src,  PIXEL *pix) {
  pix->r = src->r;
  pix->g = src->g;
  pix->b = src->b;
}

inline unsigned char FLT2UCHAR(float x) {
  return ((unsigned char) MIN(fceil(x), 255.0));
}

class region2d {
public:

  region2d() {};

  int mtrx_sub_prt_2op( struct mtrx* dest, MTRX_IT m1_start_w, 
			MTRX_IT m1_start_h,
			struct mtrx* src, MTRX_IT m2_start_w, 
			MTRX_IT m2_start_h,
			MTRX_IT width, MTRX_IT height,
			MTRXoprtFunc func) {
    MTRX_TYPE* data1end,*line1end,*data1,*data2;
    MTRX_IT line1inc, data1inc, data2inc, ret=0;
    
    data1    = dest->data;
    data2    = src->data;
    data1end = ((m1_start_h+height-1))*dest->width + (m1_start_w + width) + dest->data;
    line1end = dest->data +width+(((m1_start_h)*(dest->width)) + m1_start_w );
    data1inc = (dest->width - width);
    data2inc = (src->width - width);
    line1inc = dest->width;
    data1    += (((m1_start_h)*(dest->width)) + m1_start_w );
    data2    += (((m2_start_h)*(src->width)) + m2_start_w);
    for(;data1< data1end; ) {
      for(; data1<line1end; ){
	ret += func((data1), (data2 ), (void *) this);
	data1++;
	data2++;
      }
      data2 += data2inc;
      data1 += data1inc;
      line1end += line1inc;
    }
    return ret;
  }
  
  int mtrx_sub_prt_windowAverage( struct mtrx* src, 
				  MTRX_IT x, MTRX_IT y,
				  MTRX_IT width, MTRX_IT height, 
				  realRGB *meanrgb) {
    MTRX_TYPE *pix;
    
    meanrgb->r = 0.0;
    meanrgb->g = 0.0;
    meanrgb->b = 0.0;
    
    realRGB sumrgb;
    // float total;

    sumrgb.r = 0.0;
    sumrgb.g = 0.0;
    sumrgb.b = 0.0;

    // ignore boundary for now
    // center = src->data + y*src->width + x;

    int i, j;
    for(i=-1;i<=1; i++) {
      pix = src->data + (y+i)*src->width + (x-1);
      for(j=-1;j<=1; j++){	
	// total = ((float) pix->r) + ((float)pix->g) + ((float)pix->b);
	sumrgb.r += ((float)pix->r);// / total;
	sumrgb.g += ((float)pix->g);// / total;
	sumrgb.b += ((float)pix->b);// / total;
	pix++;
      }
    }

    float k = (float) (width*height);
    meanrgb->r = sumrgb.r / k;
    meanrgb->g = sumrgb.g / k;
    meanrgb->b = sumrgb.b / k;
    return 0;
  }

};


#endif
